"""
Generic HTTP transport base classes and utilities.
"""

import dataclasses
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping
from typing import Optional


@dataclasses.dataclass
class Response:
    content: bytes
    status_code: int
    headers: Mapping[str, str]


class RequestBase(metaclass=ABCMeta):
    """
    Abstract base class for synchronous HTTP transports.
    """

    @abstractmethod
    def __call__(
        self,
        url: str,
        body: Optional[bytes] = None,
        method: Optional[str] = None,
        headers: Optional[Mapping[str, str]] = None,
    ) -> Response:
        """
        Perform a HTTP request.

        Args:
            url: URL of resource to request.
            body: Body of request. Defaults to an empty body.
            method: HTTP method for request. Defaults to 'GET'.
            headers: Map of headers to set on the request. Defaults to an empty mapping.

        Returns:
            The response from the resource server.

        Raises:
            federatedidentity.exceptions.TransportError: on any transport error such as DNS
                resolution failure. Note that error status codes from the server do not raise.
        """


class AsyncRequestBase(metaclass=ABCMeta):
    """
    Abstract base class for asynchronous HTTP transports.
    """

    @abstractmethod
    async def __call__(
        self,
        url: str,
        body: Optional[bytes] = None,
        method: Optional[str] = None,
        headers: Optional[Mapping[str, str]] = None,
    ) -> Response:
        """
        Perform a HTTP request.

        Args:
            url: URL of resource to request.
            body: Body of request. Defaults to an empty body.
            method: HTTP method for request. Defaults to 'GET'.
            headers: Map of headers to set on the request. Defaults to an empty mapping.

        Returns:
            The response from the resource server.

        Raises:
            federatedidentity.exceptions.TransportError: on any transport error such as DNS
                resolution failure. Note that error status codes from the server do not raise.
        """



"""
Utility functions for constructing claims verifiers suitable for passing to
[verify_id_token][federatedidentity.verify_id_token].

"""

from collections.abc import Container, Iterable
from typing import Any

from ._verify import ClaimVerifier, _verify_claims
from .exceptions import InvalidClaimsError


def all_claims_present(claim_names: Iterable[str]) -> ClaimVerifier:
    """
    Verifies that all claims in `required_claims` are present.

    Arguments:
        claim_names: Iterable of claim names.

    Returns:
        A claims verifier.
    """

    def verify(claims: dict[str, Any]):
        missing_claims = set(claim_names) - set(claims.keys())
        if len(missing_claims) > 0:
            raise InvalidClaimsError(
                "Required claims "
                f"{', '.join(repr(c) for c in missing_claims)} not present in token"
            )

    return verify


def only_for_issuers(
    issuers: Container[str], required_claims: Iterable[ClaimVerifier]
) -> ClaimVerifier:
    """
    Apply claim verifiers only for a particular set of issuers.

    Arguments:
        issuers: Issuer names which should have the claims verifiers in `required_claims` applied.
        required_claims: Iterable of claim verifiers to run if the token matches one of the issuers
            in `issuers`.

    Returns:
        A claims verifier.
    """

    def verify(claims: dict[str, Any]):
        if claims["iss"] not in issuers:
            return
        _verify_claims(claims, required_claims)

    return verify



class FederatedIdentityError(RuntimeError):
    "Base class for all errors raised by the federatedidentity module."


class InvalidIssuerError(FederatedIdentityError):
    "The issuer claim in the JWT was not correctly formed."


class InvalidJWKSUrlError(FederatedIdentityError):
    "The JWKS URL in the OIDC discovery document was not correctly formed."


class InvalidOIDCDiscoveryDocumentError(FederatedIdentityError):
    "The OIDC discovery document was malformed."


class InvalidTokenError(FederatedIdentityError):
    "The token was malformed or could not be validated against the issuer public key."


class TransportError(FederatedIdentityError):
    "There was an error fetching a URL."


class InvalidClaimsError(FederatedIdentityError):
    "The claims in the token did not match policy."
